import * as THREE from "three";
import {
  createAvatarIK,
  isAnimationDataLoaded,
  preLoadAnimationData,
  createAvatar,
  registerSyncAvatarHeadAndCamera,
  setNonVRCameraMode,
  addMirrorHUD,
  Lipsync,
  SimpleBoundingBoxCollider,
} from "three-avatar";
import { setupScene, createTransformControls } from "./setup";
import { PlayerController } from "./player-controller";

let _avatar;
let _avatarIK;
let _collisionBoxes = [];
let _collisionObjects = [];
let _interactableObjects = [];
let _teleportTargetObjects = [];
let _isFPS = false;
let _bc;

export const add_other = ()=>{
  //initialLoadは済んでる前提
  
}

export const main = async (initialLoad /* :?()=>Promise<void>*/) => {
  let lipSync;
  const ctx = setupScene((dt) => {
    
    //2024-03-03 for nocore multi
    if("tick_others" in window){
      window.tick_others(dt);
    }
    if("labelRenderer" in window){
      labelRenderer.render(scene, camera);
    }
  
    _avatar?.tick(dt);
    _avatarIK?.tick(dt);

    if (ctx?.renderer.xr?.enabled) {
      _avatar?.headSync(ctx?.camera.rotation);
    }
    if (lipSync) {
      _avatar?.lipSync(...lipSync.update());
    }
    playerController?.tick(dt);
  }, true);
  _teleportTargetObjects.push(ctx.ground);

  const { scene, cameraContainer, renderer, camera } = ctx;
  const playerObj = new THREE.Group();
  playerObj.name = "playerObj";
  playerObj.add(cameraContainer);
  scene.add(playerObj);

  window.mycamera = camera;
  //
  window.myplayer = playerObj;
  window._collisionBoxes = _collisionBoxes;
  _bc = new SimpleBoundingBoxCollider(playerObj, () => _collisionBoxes);

  const playerController = new PlayerController(
    playerObj,
    cameraContainer,
    cameraContainer,
    scene,
    renderer,
    camera,
    {
      moveTo: _bc.moveTo.bind(_bc),
      getCollisionObjects: () => _collisionObjects,
      getInteractableObjects: () => _interactableObjects,
      getTeleportTargetObjects: () => _teleportTargetObjects,
    }
  );

  if (!initialLoad) {
    setupVR(ctx, playerController);
  } else {
    setTimeout(async () => {
      await initialLoad();
      setupVR(ctx, playerController);
    });
  }
  window.the_playerController = playerController;
  
  const setTestIKEnabled = setupTestIK(ctx, playerObj, playerController);
  let mirrorHUD;
  let lastUrl;
  let lastAnimationMap;
  let isLowSpecMode = false;
  return {
    scene,
    collisionObjects: _collisionObjects,
    interactableObjects: _interactableObjects,
    teleportTargetObjects: _teleportTargetObjects,
    getAvatar: () => _avatar,
    setLowSpecMode: (enabled) => {
      enabled = !!enabled;
      if (enabled === isLowSpecMode) {
        return;
      }
      isLowSpecMode = enabled;
      if (lastUrl) {
        return changeAvatar(
          camera,
          lastUrl,
          lastAnimationMap,
          playerObj,
          renderer,
          isLowSpecMode
        );
      }
    },
    changeAvatar: (url, animationMap) => {
      lastUrl = url;
      lastAnimationMap = animationMap;
      return changeAvatar(
        camera,
        url,
        animationMap,
        playerObj,
        renderer,
        isLowSpecMode
      );
    },
    setLipsyncEnabled: async (enabled) => {
      if (enabled) {
        const ms = await navigator.mediaDevices.getUserMedia({
          audio: true,
        });
        lipSync = new Lipsync(THREE.AudioContext.getContext(), ms);
      } else {
        lipSync = undefined;
      }
    },
    setTestIKEnabled,
    setFPSMode: (enabled) => {
      _isFPS = enabled;
      updateNonVrCameraMode(_isFPS, _avatar, ctx.camera);
    },
    showMirrorHUD: async (enabled) => {
      mirrorHUD?.removeFromParent();
      mirrorHUD?.dispose();
      if (enabled) {
        while (!_avatar) {
          await new Promise((resolve) => setTimeout(resolve));
        }
        mirrorHUD = addMirrorHUD(_avatar, playerObj, {
          xr: renderer.xr,
        });
      } else {
        mirrorHUD = undefined;
      }
    },
  };
};

//なんか判定座標おかしいときがあるので強制修正処理
window.fix_collision = async () =>{
  _collisionBoxes.length = 0;
  [..._collisionObjects, ..._teleportTargetObjects]
  .map((el) => {
    el.traverse((c) => {
      if (!c.isMesh) {
        return;
      }
      _collisionBoxes.push(new THREE.Box3().setFromObject(c));
    });
  });
};

const changeAvatar = async (
  camera,
  url,
  animationMap,
  playerObj,
  renderer,
  isLowSpecMode
) => {
  _avatar?.dispose();

  if (!isAnimationDataLoaded()) {
    await preLoadAnimationData(animationMap);
  }

  //urlをArrayBufferも受け付けるようにした
  let arbuff = null;
  if(typeof url =="string"){
    let resp = await fetch(url);
    arbuff = await resp.arrayBuffer();
  }else{
    arbuff = url;
  }
  const avatarData = new Uint8Array(arbuff);

  _avatar = await createAvatar(avatarData, renderer, false, {
    isInvisibleFirstPerson: false,
    isLowSpecMode,
  });
  window._avatar = _avatar;
  _avatar.object3D.name = "myAvatar";
  playerObj.add(_avatar.object3D);
  window._teleportTargetObjects = _teleportTargetObjects;
  _bc.setup(_avatar);
  _collisionBoxes.length = 0;
  [..._collisionObjects, ..._teleportTargetObjects]
  .map((el) => {
    el.traverse((c) => {
      if (!c.isMesh) {
        return;
      }
      _collisionBoxes.push(new THREE.Box3().setFromObject(c));
    });
  });
  

  if (renderer.xr.isPresenting) {
    const getCameras = () =>
      [camera, renderer.xr?.getCamera()].filter((v) => !!v);
    _avatar.setFirstPersonMode(getCameras());
  } else {
    updateNonVrCameraMode(_isFPS, _avatar, camera);
  }
};

const setupVR = ({ camera, renderer }, playerController) => {
  const getCameras = () =>
    [camera, renderer.xr?.getCamera()].filter((v) => !!v);

window.enable_handtrack = async function(){
  window.the_renderer = renderer;
  await _avatar.setIKMode(true);

  let rhand = new THREE.Object3D();
  let lhand = new THREE.Object3D();
  _avatar._object3D.add(rhand);
  _avatar._object3D.add(lhand);
  setInterval(function(){
    rhand.position.copy(renderer.xr.getController(0).position);
    rhand.quaternion.copy(renderer.xr.getController(0).quaternion);
    lhand.position.copy(renderer.xr.getController(1).position);
    lhand.quaternion.copy(renderer.xr.getController(1).quaternion);
  },100);


  _avatarIK = createAvatarIK(
    _avatar,
    {
      left: () => playerController.xrController.handHolder.leftHand,
      right: () => playerController.xrController.handHolder.rightHand,
    },
    { isDebug: true }
  );
}


  registerSyncAvatarHeadAndCamera(
    renderer.xr,
    camera,
    camera,
    camera.parent,
    () => _avatar,
    {
      onVR: async () => {
        playerController.isVR = true;
        await _avatar.setIKMode(true);
        //_avatar.setFirstPersonMode(getCameras());
        _avatarIK?.dispose();
        
/*
        _avatarIK = createAvatarIK(
          _avatar,
          {
            left: () => playerController.xrController.handHolder.leftHand,
            right: () => playerController.xrController.handHolder.rightHand,
          },
          { isDebug: true }
        );
*/
        
        window._debugAik = _avatarIK;
      },
      onNonVR: async () => {
        await _avatar.setIKMode(false);
        updateNonVrCameraMode(_isFPS, _avatar, camera);
        _avatarIK?.dispose();
        _avatarIK = undefined;

        playerController.isVR = false;
      },
    }
  );
};
const updateNonVrCameraMode = (isFPS, avatar, camera) => {
  setNonVRCameraMode(camera, camera.parent, avatar, isFPS);
};
const setupTestIK = (
  { renderer, camera, scene },
  playerObj,
  playerController
) => {
  const createTarget = (pos) => {
    const tc = createTransformControls(
      camera,
      renderer.domElement,
      (enabled) => {
        playerController.enabled = !enabled;
      }
    );
    scene.add(tc);
    const mesh = new THREE.Mesh(
      new THREE.BoxGeometry(0.01, 0.01, 0.01),
      new THREE.MeshNormalMaterial()
    );
    mesh.position.set(...pos);
    playerObj.add(mesh);
    tc.attach(mesh);
    tc.visible = tc.enabled = mesh.visible = false;
    return { tc, mesh };
  };
  const targets = [
    createTarget([0.26, 0.7, -0.26]),
    createTarget([-0.26, 0.7, -0.26]),
  ];

  const f = async (enabled) => {
    targets.forEach((v) => {
      v.tc.visible = v.tc.enabled = v.mesh.visible = enabled;
    });
    await _avatar.setIKMode(enabled);
    _avatarIK?.dispose();
    if (enabled) {
      _avatarIK = createAvatarIK(
        _avatar,
        { right: () => targets[0].mesh, left: () => targets[1].mesh },
        { isDebug: true }
      );
    } else {
      _avatarIK = undefined;
    }
    window._debugAik = _avatarIK;
  };
  return f;
};
